pub(crate) use builtin_argument_shadowing::*;
pub(crate) use builtin_attribute_shadowing::*;
pub(crate) use builtin_variable_shadowing::*;

mod builtin_argument_shadowing;
mod builtin_attribute_shadowing;
mod builtin_variable_shadowing;
