//! This file is generated by `scripts/generate_known_standard_library.py`
use once_cell::sync::Lazy;
use rustc_hash::{FxHashMap, FxHashSet};

// See: https://pycqa.github.io/isort/docs/configuration/options.html#known-standard-library
pub static KNOWN_STANDARD_LIBRARY: Lazy<FxHashMap<(u32, u32), FxHashSet<&'static str>>> =
    Lazy::new(|| {
        FxHashMap::from_iter([
            (
                (3, 7),
                FxHashSet::from_iter([
                    "_ast",
                    "_dummy_thread",
                    "_thread",
                    "abc",
                    "aifc",
                    "argparse",
                    "array",
                    "ast",
                    "asynchat",
                    "asyncio",
                    "asyncore",
                    "atexit",
                    "audioop",
                    "base64",
                    "bdb",
                    "binascii",
                    "binhex",
                    "bisect",
                    "builtins",
                    "bz2",
                    "cProfile",
                    "calendar",
                    "cgi",
                    "cgitb",
                    "chunk",
                    "cmath",
                    "cmd",
                    "code",
                    "codecs",
                    "codeop",
                    "collections",
                    "colorsys",
                    "compileall",
                    "concurrent",
                    "configparser",
                    "contextlib",
                    "contextvars",
                    "copy",
                    "copyreg",
                    "crypt",
                    "csv",
                    "ctypes",
                    "curses",
                    "dataclasses",
                    "datetime",
                    "dbm",
                    "decimal",
                    "difflib",
                    "dis",
                    "distutils",
                    "doctest",
                    "dummy_threading",
                    "email",
                    "encodings",
                    "ensurepip",
                    "enum",
                    "errno",
                    "faulthandler",
                    "fcntl",
                    "filecmp",
                    "fileinput",
                    "fnmatch",
                    "formatter",
                    "fractions",
                    "ftplib",
                    "functools",
                    "gc",
                    "getopt",
                    "getpass",
                    "gettext",
                    "glob",
                    "grp",
                    "gzip",
                    "hashlib",
                    "heapq",
                    "hmac",
                    "html",
                    "http",
                    "imaplib",
                    "imghdr",
                    "imp",
                    "importlib",
                    "inspect",
                    "io",
                    "ipaddress",
                    "itertools",
                    "json",
                    "keyword",
                    "lib2to3",
                    "linecache",
                    "locale",
                    "logging",
                    "lzma",
                    "macpath",
                    "mailbox",
                    "mailcap",
                    "marshal",
                    "math",
                    "mimetypes",
                    "mmap",
                    "modulefinder",
                    "msilib",
                    "msvcrt",
                    "multiprocessing",
                    "netrc",
                    "nis",
                    "nntplib",
                    "ntpath",
                    "numbers",
                    "operator",
                    "optparse",
                    "os",
                    "ossaudiodev",
                    "parser",
                    "pathlib",
                    "pdb",
                    "pickle",
                    "pickletools",
                    "pipes",
                    "pkgutil",
                    "platform",
                    "plistlib",
                    "poplib",
                    "posix",
                    "posixpath",
                    "pprint",
                    "profile",
                    "pstats",
                    "pty",
                    "pwd",
                    "py_compile",
                    "pyclbr",
                    "pydoc",
                    "queue",
                    "quopri",
                    "random",
                    "re",
                    "readline",
                    "reprlib",
                    "resource",
                    "rlcompleter",
                    "runpy",
                    "sched",
                    "secrets",
                    "select",
                    "selectors",
                    "shelve",
                    "shlex",
                    "shutil",
                    "signal",
                    "site",
                    "smtpd",
                    "smtplib",
                    "sndhdr",
                    "socket",
                    "socketserver",
                    "spwd",
                    "sqlite3",
                    "sre",
                    "sre_compile",
                    "sre_constants",
                    "sre_parse",
                    "ssl",
                    "stat",
                    "statistics",
                    "string",
                    "stringprep",
                    "struct",
                    "subprocess",
                    "sunau",
                    "symbol",
                    "symtable",
                    "sys",
                    "sysconfig",
                    "syslog",
                    "tabnanny",
                    "tarfile",
                    "telnetlib",
                    "tempfile",
                    "termios",
                    "test",
                    "textwrap",
                    "threading",
                    "time",
                    "timeit",
                    "tkinter",
                    "token",
                    "tokenize",
                    "trace",
                    "traceback",
                    "tracemalloc",
                    "tty",
                    "turtle",
                    "turtledemo",
                    "types",
                    "typing",
                    "unicodedata",
                    "unittest",
                    "urllib",
                    "uu",
                    "uuid",
                    "venv",
                    "warnings",
                    "wave",
                    "weakref",
                    "webbrowser",
                    "winreg",
                    "winsound",
                    "wsgiref",
                    "xdrlib",
                    "xml",
                    "xmlrpc",
                    "zipapp",
                    "zipfile",
                    "zipimport",
                    "zlib",
                ]),
            ),
            (
                (3, 8),
                FxHashSet::from_iter([
                    "_ast",
                    "_dummy_thread",
                    "_thread",
                    "abc",
                    "aifc",
                    "argparse",
                    "array",
                    "ast",
                    "asynchat",
                    "asyncio",
                    "asyncore",
                    "atexit",
                    "audioop",
                    "base64",
                    "bdb",
                    "binascii",
                    "binhex",
                    "bisect",
                    "builtins",
                    "bz2",
                    "cProfile",
                    "calendar",
                    "cgi",
                    "cgitb",
                    "chunk",
                    "cmath",
                    "cmd",
                    "code",
                    "codecs",
                    "codeop",
                    "collections",
                    "colorsys",
                    "compileall",
                    "concurrent",
                    "configparser",
                    "contextlib",
                    "contextvars",
                    "copy",
                    "copyreg",
                    "crypt",
                    "csv",
                    "ctypes",
                    "curses",
                    "dataclasses",
                    "datetime",
                    "dbm",
                    "decimal",
                    "difflib",
                    "dis",
                    "distutils",
                    "doctest",
                    "dummy_threading",
                    "email",
                    "encodings",
                    "ensurepip",
                    "enum",
                    "errno",
                    "faulthandler",
                    "fcntl",
                    "filecmp",
                    "fileinput",
                    "fnmatch",
                    "formatter",
                    "fractions",
                    "ftplib",
                    "functools",
                    "gc",
                    "getopt",
                    "getpass",
                    "gettext",
                    "glob",
                    "grp",
                    "gzip",
                    "hashlib",
                    "heapq",
                    "hmac",
                    "html",
                    "http",
                    "imaplib",
                    "imghdr",
                    "imp",
                    "importlib",
                    "inspect",
                    "io",
                    "ipaddress",
                    "itertools",
                    "json",
                    "keyword",
                    "lib2to3",
                    "linecache",
                    "locale",
                    "logging",
                    "lzma",
                    "mailbox",
                    "mailcap",
                    "marshal",
                    "math",
                    "mimetypes",
                    "mmap",
                    "modulefinder",
                    "msilib",
                    "msvcrt",
                    "multiprocessing",
                    "netrc",
                    "nis",
                    "nntplib",
                    "ntpath",
                    "numbers",
                    "operator",
                    "optparse",
                    "os",
                    "ossaudiodev",
                    "parser",
                    "pathlib",
                    "pdb",
                    "pickle",
                    "pickletools",
                    "pipes",
                    "pkgutil",
                    "platform",
                    "plistlib",
                    "poplib",
                    "posix",
                    "posixpath",
                    "pprint",
                    "profile",
                    "pstats",
                    "pty",
                    "pwd",
                    "py_compile",
                    "pyclbr",
                    "pydoc",
                    "queue",
                    "quopri",
                    "random",
                    "re",
                    "readline",
                    "reprlib",
                    "resource",
                    "rlcompleter",
                    "runpy",
                    "sched",
                    "secrets",
                    "select",
                    "selectors",
                    "shelve",
                    "shlex",
                    "shutil",
                    "signal",
                    "site",
                    "smtpd",
                    "smtplib",
                    "sndhdr",
                    "socket",
                    "socketserver",
                    "spwd",
                    "sqlite3",
                    "sre",
                    "sre_compile",
                    "sre_constants",
                    "sre_parse",
                    "ssl",
                    "stat",
                    "statistics",
                    "string",
                    "stringprep",
                    "struct",
                    "subprocess",
                    "sunau",
                    "symbol",
                    "symtable",
                    "sys",
                    "sysconfig",
                    "syslog",
                    "tabnanny",
                    "tarfile",
                    "telnetlib",
                    "tempfile",
                    "termios",
                    "test",
                    "textwrap",
                    "threading",
                    "time",
                    "timeit",
                    "tkinter",
                    "token",
                    "tokenize",
                    "trace",
                    "traceback",
                    "tracemalloc",
                    "tty",
                    "turtle",
                    "turtledemo",
                    "types",
                    "typing",
                    "unicodedata",
                    "unittest",
                    "urllib",
                    "uu",
                    "uuid",
                    "venv",
                    "warnings",
                    "wave",
                    "weakref",
                    "webbrowser",
                    "winreg",
                    "winsound",
                    "wsgiref",
                    "xdrlib",
                    "xml",
                    "xmlrpc",
                    "zipapp",
                    "zipfile",
                    "zipimport",
                    "zlib",
                ]),
            ),
            (
                (3, 9),
                FxHashSet::from_iter([
                    "_ast",
                    "_thread",
                    "abc",
                    "aifc",
                    "argparse",
                    "array",
                    "ast",
                    "asynchat",
                    "asyncio",
                    "asyncore",
                    "atexit",
                    "audioop",
                    "base64",
                    "bdb",
                    "binascii",
                    "binhex",
                    "bisect",
                    "builtins",
                    "bz2",
                    "cProfile",
                    "calendar",
                    "cgi",
                    "cgitb",
                    "chunk",
                    "cmath",
                    "cmd",
                    "code",
                    "codecs",
                    "codeop",
                    "collections",
                    "colorsys",
                    "compileall",
                    "concurrent",
                    "configparser",
                    "contextlib",
                    "contextvars",
                    "copy",
                    "copyreg",
                    "crypt",
                    "csv",
                    "ctypes",
                    "curses",
                    "dataclasses",
                    "datetime",
                    "dbm",
                    "decimal",
                    "difflib",
                    "dis",
                    "distutils",
                    "doctest",
                    "email",
                    "encodings",
                    "ensurepip",
                    "enum",
                    "errno",
                    "faulthandler",
                    "fcntl",
                    "filecmp",
                    "fileinput",
                    "fnmatch",
                    "formatter",
                    "fractions",
                    "ftplib",
                    "functools",
                    "gc",
                    "getopt",
                    "getpass",
                    "gettext",
                    "glob",
                    "graphlib",
                    "grp",
                    "gzip",
                    "hashlib",
                    "heapq",
                    "hmac",
                    "html",
                    "http",
                    "imaplib",
                    "imghdr",
                    "imp",
                    "importlib",
                    "inspect",
                    "io",
                    "ipaddress",
                    "itertools",
                    "json",
                    "keyword",
                    "lib2to3",
                    "linecache",
                    "locale",
                    "logging",
                    "lzma",
                    "mailbox",
                    "mailcap",
                    "marshal",
                    "math",
                    "mimetypes",
                    "mmap",
                    "modulefinder",
                    "msilib",
                    "msvcrt",
                    "multiprocessing",
                    "netrc",
                    "nis",
                    "nntplib",
                    "ntpath",
                    "numbers",
                    "operator",
                    "optparse",
                    "os",
                    "ossaudiodev",
                    "parser",
                    "pathlib",
                    "pdb",
                    "pickle",
                    "pickletools",
                    "pipes",
                    "pkgutil",
                    "platform",
                    "plistlib",
                    "poplib",
                    "posix",
                    "posixpath",
                    "pprint",
                    "profile",
                    "pstats",
                    "pty",
                    "pwd",
                    "py_compile",
                    "pyclbr",
                    "pydoc",
                    "queue",
                    "quopri",
                    "random",
                    "re",
                    "readline",
                    "reprlib",
                    "resource",
                    "rlcompleter",
                    "runpy",
                    "sched",
                    "secrets",
                    "select",
                    "selectors",
                    "shelve",
                    "shlex",
                    "shutil",
                    "signal",
                    "site",
                    "smtpd",
                    "smtplib",
                    "sndhdr",
                    "socket",
                    "socketserver",
                    "spwd",
                    "sqlite3",
                    "sre",
                    "sre_compile",
                    "sre_constants",
                    "sre_parse",
                    "ssl",
                    "stat",
                    "statistics",
                    "string",
                    "stringprep",
                    "struct",
                    "subprocess",
                    "sunau",
                    "symbol",
                    "symtable",
                    "sys",
                    "sysconfig",
                    "syslog",
                    "tabnanny",
                    "tarfile",
                    "telnetlib",
                    "tempfile",
                    "termios",
                    "test",
                    "textwrap",
                    "threading",
                    "time",
                    "timeit",
                    "tkinter",
                    "token",
                    "tokenize",
                    "trace",
                    "traceback",
                    "tracemalloc",
                    "tty",
                    "turtle",
                    "turtledemo",
                    "types",
                    "typing",
                    "unicodedata",
                    "unittest",
                    "urllib",
                    "uu",
                    "uuid",
                    "venv",
                    "warnings",
                    "wave",
                    "weakref",
                    "webbrowser",
                    "winreg",
                    "winsound",
                    "wsgiref",
                    "xdrlib",
                    "xml",
                    "xmlrpc",
                    "zipapp",
                    "zipfile",
                    "zipimport",
                    "zlib",
                    "zoneinfo",
                ]),
            ),
            (
                (3, 10),
                FxHashSet::from_iter([
                    "_ast",
                    "_thread",
                    "abc",
                    "aifc",
                    "argparse",
                    "array",
                    "ast",
                    "asynchat",
                    "asyncio",
                    "asyncore",
                    "atexit",
                    "audioop",
                    "base64",
                    "bdb",
                    "binascii",
                    "binhex",
                    "bisect",
                    "builtins",
                    "bz2",
                    "cProfile",
                    "calendar",
                    "cgi",
                    "cgitb",
                    "chunk",
                    "cmath",
                    "cmd",
                    "code",
                    "codecs",
                    "codeop",
                    "collections",
                    "colorsys",
                    "compileall",
                    "concurrent",
                    "configparser",
                    "contextlib",
                    "contextvars",
                    "copy",
                    "copyreg",
                    "crypt",
                    "csv",
                    "ctypes",
                    "curses",
                    "dataclasses",
                    "datetime",
                    "dbm",
                    "decimal",
                    "difflib",
                    "dis",
                    "distutils",
                    "doctest",
                    "email",
                    "encodings",
                    "ensurepip",
                    "enum",
                    "errno",
                    "faulthandler",
                    "fcntl",
                    "filecmp",
                    "fileinput",
                    "fnmatch",
                    "fractions",
                    "ftplib",
                    "functools",
                    "gc",
                    "getopt",
                    "getpass",
                    "gettext",
                    "glob",
                    "graphlib",
                    "grp",
                    "gzip",
                    "hashlib",
                    "heapq",
                    "hmac",
                    "html",
                    "http",
                    "idlelib",
                    "imaplib",
                    "imghdr",
                    "imp",
                    "importlib",
                    "inspect",
                    "io",
                    "ipaddress",
                    "itertools",
                    "json",
                    "keyword",
                    "lib2to3",
                    "linecache",
                    "locale",
                    "logging",
                    "lzma",
                    "mailbox",
                    "mailcap",
                    "marshal",
                    "math",
                    "mimetypes",
                    "mmap",
                    "modulefinder",
                    "msilib",
                    "msvcrt",
                    "multiprocessing",
                    "netrc",
                    "nis",
                    "nntplib",
                    "ntpath",
                    "numbers",
                    "operator",
                    "optparse",
                    "os",
                    "ossaudiodev",
                    "pathlib",
                    "pdb",
                    "pickle",
                    "pickletools",
                    "pipes",
                    "pkgutil",
                    "platform",
                    "plistlib",
                    "poplib",
                    "posix",
                    "posixpath",
                    "pprint",
                    "profile",
                    "pstats",
                    "pty",
                    "pwd",
                    "py_compile",
                    "pyclbr",
                    "pydoc",
                    "queue",
                    "quopri",
                    "random",
                    "re",
                    "readline",
                    "reprlib",
                    "resource",
                    "rlcompleter",
                    "runpy",
                    "sched",
                    "secrets",
                    "select",
                    "selectors",
                    "shelve",
                    "shlex",
                    "shutil",
                    "signal",
                    "site",
                    "smtpd",
                    "smtplib",
                    "sndhdr",
                    "socket",
                    "socketserver",
                    "spwd",
                    "sqlite3",
                    "sre",
                    "sre_compile",
                    "sre_constants",
                    "sre_parse",
                    "ssl",
                    "stat",
                    "statistics",
                    "string",
                    "stringprep",
                    "struct",
                    "subprocess",
                    "sunau",
                    "symtable",
                    "sys",
                    "sysconfig",
                    "syslog",
                    "tabnanny",
                    "tarfile",
                    "telnetlib",
                    "tempfile",
                    "termios",
                    "test",
                    "textwrap",
                    "threading",
                    "time",
                    "timeit",
                    "tkinter",
                    "token",
                    "tokenize",
                    "trace",
                    "traceback",
                    "tracemalloc",
                    "tty",
                    "turtle",
                    "turtledemo",
                    "types",
                    "typing",
                    "unicodedata",
                    "unittest",
                    "urllib",
                    "uu",
                    "uuid",
                    "venv",
                    "warnings",
                    "wave",
                    "weakref",
                    "webbrowser",
                    "winreg",
                    "winsound",
                    "wsgiref",
                    "xdrlib",
                    "xml",
                    "xmlrpc",
                    "zipapp",
                    "zipfile",
                    "zipimport",
                    "zlib",
                    "zoneinfo",
                ]),
            ),
            (
                (3, 11),
                FxHashSet::from_iter([
                    "_ast",
                    "_thread",
                    "abc",
                    "aifc",
                    "argparse",
                    "array",
                    "ast",
                    "asynchat",
                    "asyncio",
                    "asyncore",
                    "atexit",
                    "audioop",
                    "base64",
                    "bdb",
                    "binascii",
                    "bisect",
                    "builtins",
                    "bz2",
                    "cProfile",
                    "calendar",
                    "cgi",
                    "cgitb",
                    "chunk",
                    "cmath",
                    "cmd",
                    "code",
                    "codecs",
                    "codeop",
                    "collections",
                    "colorsys",
                    "compileall",
                    "concurrent",
                    "configparser",
                    "contextlib",
                    "contextvars",
                    "copy",
                    "copyreg",
                    "crypt",
                    "csv",
                    "ctypes",
                    "curses",
                    "dataclasses",
                    "datetime",
                    "dbm",
                    "decimal",
                    "difflib",
                    "dis",
                    "distutils",
                    "doctest",
                    "email",
                    "encodings",
                    "ensurepip",
                    "enum",
                    "errno",
                    "faulthandler",
                    "fcntl",
                    "filecmp",
                    "fileinput",
                    "fnmatch",
                    "fractions",
                    "ftplib",
                    "functools",
                    "gc",
                    "getopt",
                    "getpass",
                    "gettext",
                    "glob",
                    "graphlib",
                    "grp",
                    "gzip",
                    "hashlib",
                    "heapq",
                    "hmac",
                    "html",
                    "http",
                    "idlelib",
                    "imaplib",
                    "imghdr",
                    "imp",
                    "importlib",
                    "inspect",
                    "io",
                    "ipaddress",
                    "itertools",
                    "json",
                    "keyword",
                    "lib2to3",
                    "linecache",
                    "locale",
                    "logging",
                    "lzma",
                    "mailbox",
                    "mailcap",
                    "marshal",
                    "math",
                    "mimetypes",
                    "mmap",
                    "modulefinder",
                    "msilib",
                    "msvcrt",
                    "multiprocessing",
                    "netrc",
                    "nis",
                    "nntplib",
                    "ntpath",
                    "numbers",
                    "operator",
                    "optparse",
                    "os",
                    "ossaudiodev",
                    "pathlib",
                    "pdb",
                    "pickle",
                    "pickletools",
                    "pipes",
                    "pkgutil",
                    "platform",
                    "plistlib",
                    "poplib",
                    "posix",
                    "posixpath",
                    "pprint",
                    "profile",
                    "pstats",
                    "pty",
                    "pwd",
                    "py_compile",
                    "pyclbr",
                    "pydoc",
                    "queue",
                    "quopri",
                    "random",
                    "re",
                    "readline",
                    "reprlib",
                    "resource",
                    "rlcompleter",
                    "runpy",
                    "sched",
                    "secrets",
                    "select",
                    "selectors",
                    "shelve",
                    "shlex",
                    "shutil",
                    "signal",
                    "site",
                    "smtpd",
                    "smtplib",
                    "sndhdr",
                    "socket",
                    "socketserver",
                    "spwd",
                    "sqlite3",
                    "sre",
                    "sre_compile",
                    "sre_constants",
                    "sre_parse",
                    "ssl",
                    "stat",
                    "statistics",
                    "string",
                    "stringprep",
                    "struct",
                    "subprocess",
                    "sunau",
                    "symtable",
                    "sys",
                    "sysconfig",
                    "syslog",
                    "tabnanny",
                    "tarfile",
                    "telnetlib",
                    "tempfile",
                    "termios",
                    "test",
                    "textwrap",
                    "threading",
                    "time",
                    "timeit",
                    "tkinter",
                    "token",
                    "tokenize",
                    "tomllib",
                    "trace",
                    "traceback",
                    "tracemalloc",
                    "tty",
                    "turtle",
                    "turtledemo",
                    "types",
                    "typing",
                    "unicodedata",
                    "unittest",
                    "urllib",
                    "uu",
                    "uuid",
                    "venv",
                    "warnings",
                    "wave",
                    "weakref",
                    "webbrowser",
                    "winreg",
                    "winsound",
                    "wsgiref",
                    "xdrlib",
                    "xml",
                    "xmlrpc",
                    "zipapp",
                    "zipfile",
                    "zipimport",
                    "zlib",
                    "zoneinfo",
                ]),
            ),
        ])
    });
