if x == y: # trailing if condition
    pass # trailing `pass` comment
    # Root `if` trailing comment

# Leading elif comment
elif x < y: # trailing elif condition
    pass
    # `elif` trailing comment

# Leading else comment
else: # trailing else condition
    pass
    # `else` trailing comment


if x == y:
    if y == z:
        ...

    if a == b:
        ...
    else: # trailing comment
        ...

    # trailing else comment

# leading else if comment
elif aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa + [
    11111111111111111111111111,
    2222222222222222222222,
    3333333333
    ]:
    ...


else:
    ...
